// Read COPYING for licensing details.

#include <FL/ml.h>
#include <math.h>
#include <stdio.h>

void
mlVectorfv(MLenum pname, MLsizei cols, const MLfloat *param, MLfloat *result)
{
	switch(pname)
	{
		case ML_VECTOR_PRINT:
		{
			MLuint i;
			printf("(");
			for(i = 0; i < cols; i++) printf("%f%s", param[i], (i < (cols - 1)) ? " " : "");
			printf(")\n");
			break;
		}
		case ML_VECTOR_MODULE:
		{
			MLuint i; *result = 0;
			for(i = 0; i < cols; i++) *result += pow(param[i], 2);
			*result = sqrt(*result);
			break;
		}
		case ML_VECTOR_NORMAL:
		{
			MLfloat mod;
			mlVectorfv(ML_VECTOR_MODULE, cols, param, &mod);
			mlVector2f(ML_MULT_VECTOR_AND_SCALAR, cols, param, 1.0f / mod, result);
			break;
		}
		default: break;
	}
}

void
mlVector2f(MLenum pname, MLsizei cols, const MLfloat *param1, MLfloat param2, MLfloat *result)
{
	switch(pname)
	{
		case ML_MULT_VECTOR_AND_SCALAR:
		{
			MLuint i; 
			for(i = 0; i < cols; i++) result[i] = param1[i] * param2;
			break;
		}
		default: break;
	}
}

void
mlVector2fv(MLenum pname, MLsizei cols, const MLfloat *param1, const MLfloat *param2, MLfloat *result)
{
	switch(pname)
	{
		case ML_VECTOR_SCALAR_PRODUCT: break;
		{
			MLuint i; *result = 0;
			for(i = 0; i < cols; i++) *result += param1[i] * param2[i];
			break;
		}
		case ML_VECTOR_CROSS_PRODUCT: break;
			break;
		default: break;
	}
}

void
mlMatrixfv(MLenum pname, MLsizei rows, MLsizei cols, const MLfloat *param, MLfloat *result)
{
	switch(pname)
	{
		case ML_MATRIX_PRINT:
		{
			MLuint i, j;
			for(j = 0; j < rows; j++)
			{
				printf("(");
				for(i = 0; i < cols; i++)
					printf("%f%s", param[j + cols * i], (i < (cols - 1)) ? " " : "");
				printf(")\n");
			}
			break;
		}
		case ML_MATRIX_TRANSPOSE:
		{
			MLuint i, j;
			for(j = 0; j < rows; j++)
				for(i = 0; i < cols; i++)
					result[i + cols * j] = param[j + rows * i];
			break;
		}
		case ML_MATRIX_DETERMINANT:
		{
			MLuint i;
			if(rows == 2)
			{
				*result = param[0] * param[3] - param[1] * param[2];
			}
			else
			{
				*result = 0;
				for(i = 0; i < cols; i++)
				{
					MLfloat cofactor;
					mlMatrix3f(ML_ADJOINT, rows, cols, param, i, 0, &cofactor);
					*result += param[i] * cofactor;
				}
			}


			break;
		}
		case ML_MATRIX_ADJOINT:
		{
			MLuint i, j;
			for(j = 0; j < rows; j++)
				for(i = 0; i < cols; i++)
				{
					mlMatrix3f(ML_ADJOINT, rows, cols, param, i, j, &result[i + cols * j]);
				}
			break;
		}
		case ML_MATRIX_INVERSE:
		{
			MLfloat adj[rows * cols];
			mlMatrixfv(ML_MATRIX_ADJOINT, rows, cols, param, adj);
			MLfloat tadj[rows * cols];
			mlMatrixfv(ML_MATRIX_TRANSPOSE, rows, cols, adj, tadj);
			MLfloat det;
			mlMatrixfv(ML_MATRIX_DETERMINANT, rows, cols, param, &det);
			mlMatrix2f(ML_MULT_MATRIX_AND_SCALAR, rows, cols, tadj, 1/det, result);
			break;
		}
		default: break;
	}
}

void
mlMatrix2f(MLenum pname, MLsizei rows, MLsizei cols, const MLfloat *param1, MLfloat param2, MLfloat *result)
{
	switch(pname)
	{
		case ML_MULT_MATRIX_AND_SCALAR:
		{
			MLuint i;
			for(i = 0; i < rows * cols; i++) result[i] = param1[i] * param2;
			break;
		}
		default: break;
	}
}

void
mlMatrix3f(MLenum pname, MLsizei rows, MLsizei cols, const MLfloat *param1, MLfloat param2, MLfloat param3, MLfloat *result)
{
	switch(pname)
	{
		case ML_ADJOINT:
		{
			MLfloat a[(rows - 1) * (cols - 1)];
			MLuint i, j, i2, j2;
			for(j = 0, j2 = 0; j < rows; j++)
			{
				if(j == param3) continue;
				for(i = 0, i2 = 0; i < cols; i++)
				{
					if((MLfloat)i == param2) continue;
					a[i2 + (cols - 1) * j2] = param1[i + cols * j];
					i2++;
				}
				j2++;
			}
			MLint sign = pow(-1, param2 + param3);
			MLfloat det;
			mlMatrixfv(ML_MATRIX_DETERMINANT, rows - 1, cols - 1, a, &det);
			*result = pow(-1, param2 + param3) * det;
			break;
		}
		default: break;
	}
}

void
mlMatrix2fv(MLenum pname, MLsizei rows1, MLsizei cols1, const MLfloat *param1, MLsizei rows2, MLsizei cols2, const MLfloat *param2, MLfloat *result)
{
	switch(pname)
	{		
		case ML_ADD_MATRIX_AND_MATRIX:
		{
			int i;
			for(i = 0; i < rows1 * cols1; i++) result[i] = param1[i] + param2[i];
			break;
		}
		case ML_MULT_MATRIX_AND_MATRIX:
		{
			int i, j, k;
			for(j = 0; j < rows1; j++)
				for(i = 0; i < cols2; i++)
				{
					result[i + cols2 * j] = 0;
					for(k = 0; k < cols1; k++)
						result[i + cols2 * j] += param1[k + cols1 * j] * param2[i + cols2 * k];
				}
			break;
		}
		default: break;
	}
}
